@@ -270,7 +270,7 @@ def random_shard(shard_size, dataset_size):
 
   return _random_shard
 
-def shuffle_dataset(buffer_size, shuffle_shards=True):
+def shuffle_dataset(buffer_size, shuffle_shards=True, dataset_size=None):
   """Transformation that shuffles the dataset based on its size.
 
   Example:
@@ -285,6 +285,8 @@ def shuffle_dataset(buffer_size, shuffle_shards=True):
     shuffle_shards: When :obj:`buffer_size` is smaller than the dataset size,
       the dataset is first sharded in a random order to add another level of
       shuffling.
+    dataset_size: If the dataset size is already known, it can be passed here to
+      avoid a slower generic computation of the dataset size later.
 
   Returns:
     A ``tf.data.Dataset`` transformation.
@@ -293,12 +295,14 @@ def shuffle_dataset(buffer_size, shuffle_shards=True):
   def _shuffle(dataset):
     sample_size = buffer_size
     if sample_size < 0 or shuffle_shards:
-      dataset_size = get_dataset_size(dataset)
-      tf.get_logger().info("Training on %d examples", dataset_size)
+      total_size = dataset_size
+      if total_size is None:
+        total_size = get_dataset_size(dataset)
+      tf.get_logger().info("Training on %d examples", total_size)
       if sample_size < 0:
-        sample_size = dataset_size
-      elif sample_size < dataset_size:
-        dataset = dataset.apply(random_shard(sample_size, dataset_size))
+        sample_size = total_size
+      elif sample_size < total_size:
+        dataset = dataset.apply(random_shard(sample_size, total_size))
     dataset = dataset.shuffle(sample_size)
     return dataset
 
@@ -439,6 +443,7 @@ def training_pipeline(batch_size,
                       batch_multiplier=1,
                       batch_size_multiple=1,
                       process_fn=None,
+                      transform_fns=None,
                       length_bucket_width=None,
                       features_length_fn=None,
                       labels_length_fn=None,
@@ -448,6 +453,7 @@ def training_pipeline(batch_size,
                       num_shards=1,
                       shard_index=0,
                       num_threads=None,
+                      dataset_size=None,
                       shuffle_buffer_size=None,
                       prefetch_buffer_size=None,
                       cardinality_multiple=1):
@@ -474,6 +480,8 @@ def training_pipeline(batch_size,
     batch_size_multiple: When :obj:`batch_type` is "tokens", ensure that the
       resulting batch size is a multiple of this value.
     process_fn: The processing function to apply on each element.
+    transform_fns: List of dataset transformation functions (applied after
+      :obj:`process_fn` if defined).
     length_bucket_width: The width of the length buckets to select batch
       candidates from. ``None`` to not constrain batch formation.
     features_length_fn: A function mapping features to a sequence length.
@@ -487,6 +495,8 @@ def training_pipeline(batch_size,
       distributed setting).
     shard_index: The shard index this data pipeline should read from.
     num_threads: The number of elements processed in parallel.
+    dataset_size: If the dataset size is already known, it can be passed here to
+      avoid a slower generic computation of the dataset size later.
     shuffle_buffer_size: The number of elements from which to sample.
     prefetch_buffer_size: The number of batches to prefetch asynchronously. If
       ``None``, use an automatically tuned value.
@@ -503,6 +513,12 @@ def training_pipeline(batch_size,
     - :func:`opennmt.data.filter_irregular_batches`
     - :func:`opennmt.data.shuffle_dataset`
   """
+  if dataset_size is not None and num_shards > 1:
+    # Update dataset_size based on the shard size.
+    if isinstance(dataset_size, list):
+      dataset_size = [size // num_shards for size in dataset_size]
+    else:
+      dataset_size //= num_shards
 
   def _make_weighted_dataset(datasets, weights):
     if single_pass:
@@ -514,7 +530,7 @@ def training_pipeline(batch_size,
                        "number of data files" % (len(weights), len(datasets)))
     if num_shards > 1:
       datasets = [dataset.shard(num_shards, shard_index) for dataset in datasets]
-    weights = normalize_weights(datasets, weights=weights)
+    weights = normalize_weights(datasets, weights=weights, sizes=dataset_size)
     datasets = [dataset.repeat() for dataset in datasets]
     dataset = tf.data.experimental.sample_from_datasets(datasets, weights=weights)
     if shuffle_buffer_size is not None and shuffle_buffer_size != 0:
@@ -527,7 +543,7 @@ def training_pipeline(batch_size,
     if num_shards > 1:
       dataset = dataset.shard(num_shards, shard_index)
     if shuffle_buffer_size is not None and shuffle_buffer_size != 0:
-      dataset = dataset.apply(shuffle_dataset(shuffle_buffer_size))
+      dataset = dataset.apply(shuffle_dataset(shuffle_buffer_size, dataset_size=dataset_size))
     return dataset
 
   def _pipeline(dataset):
@@ -542,6 +558,9 @@ def training_pipeline(batch_size,
       dataset = _make_single_dataset(dataset)
     if process_fn is not None:
       dataset = dataset.map(process_fn, num_parallel_calls=num_threads or 4)
+    if transform_fns is not None:
+      for transform_fn in transform_fns:
+        dataset = dataset.apply(transform_fn)
     dataset = dataset.apply(filter_examples_by_length(
         maximum_features_length=maximum_features_length,
         maximum_labels_length=maximum_labels_length,
@@ -567,6 +586,7 @@ def training_pipeline(batch_size,
 
 def inference_pipeline(batch_size,
                        process_fn=None,
+                       transform_fns=None,
                        length_bucket_width=None,
                        length_fn=None,
                        num_threads=None,
@@ -580,6 +600,8 @@ def inference_pipeline(batch_size,
   Args:
     batch_size: The batch size to use.
     process_fn: The processing function to apply on each element.
+    transform_fns: List of dataset transformation functions (applied after
+      :obj:`process_fn` if defined).
     length_bucket_width: The width of the length buckets to select batch
       candidates from. If set, this means the inference pipeline will be
       reordered based on the examples length, the application is then
@@ -606,6 +628,9 @@ def inference_pipeline(batch_size,
   def _pipeline(dataset):
     if process_fn is not None:
       dataset = dataset.map(process_fn, num_parallel_calls=num_threads)
+    if transform_fns is not None:
+      for transform_fn in transform_fns:
+        dataset = dataset.apply(transform_fn)
     if length_bucket_width is not None and length_bucket_width > 0:
       if length_fn is None:
         raise ValueError("length_fn is required when reordering by length")
