@@ -61,6 +61,23 @@ class Inputter(tf.keras.layers.Layer):
     """
     raise NotImplementedError()
 
+  def get_dataset_size(self, data_file):
+    """Returns the dataset size.
+
+    If the inputter can efficiently compute the dataset size from a training
+    file on disk, it can optionally override this method. Otherwise, we may
+    compute the size later with a generic and slower approach (iterating over
+    the dataset instance).
+
+    Args:
+      data_file: The data file.
+
+    Returns:
+      The dataset size or ``None``.
+    """
+    _ = data_file
+    return None
+
   def make_inference_dataset(self,
                              features_file,
                              batch_size,
@@ -88,11 +105,20 @@ class Inputter(tf.keras.layers.Layer):
     See Also:
       :func:`opennmt.data.inference_pipeline`
     """
-    map_func = lambda *arg: self.make_features(element=misc.item_or_tuple(arg), training=False)
+
+    def _map_fn(*arg):
+      features = self.make_features(element=misc.item_or_tuple(arg), training=False)
+      if isinstance(features, (list, tuple)):
+        # Special case for unsupervised inputters that always return a tuple (features, labels).
+        return features[0]
+      return features
+
+    transform_fns = [lambda dataset: dataset.map(_map_fn, num_parallel_calls=num_threads or 1)]
+
     dataset = self.make_dataset(features_file, training=False)
     dataset = dataset.apply(dataset_util.inference_pipeline(
         batch_size,
-        process_fn=map_func,
+        transform_fns=transform_fns,
         length_bucket_width=length_bucket_width,
         length_fn=self.get_length,
         num_threads=num_threads,
@@ -141,6 +167,27 @@ class Inputter(tf.keras.layers.Layer):
     """
     raise NotImplementedError()
 
+  def keep_for_training(self, features, maximum_length=None):
+    """Returns ``True`` if this example should be kept for training.
+
+    Args:
+      features: A dictionary of ``tf.Tensor``.
+      maximum_length: The maximum length used for training.
+
+    Returns:
+      A boolean.
+    """
+    if isinstance(features, (list, tuple)):
+      # Special case for unsupervised inputters that always return a tuple (features, labels).
+      features = features[0]
+    length = self.get_length(features)
+    if length is None:
+      return True
+    is_valid = tf.greater(length, 0)
+    if maximum_length is not None:
+      is_valid = tf.logical_and(is_valid, tf.less_equal(length, maximum_length))
+    return is_valid
+
   def call(self, features, training=None):  # pylint: disable=arguments-differ
     """Creates the model input from the features (e.g. word embeddings).
 
@@ -219,6 +266,13 @@ class MultiInputter(Inputter):
   def make_dataset(self, data_file, training=None):
     raise NotImplementedError()
 
+  def get_dataset_size(self, data_file):
+    for inputter, data in zip(self.inputters, data_file):
+      size = inputter.get_dataset_size(data)
+      if size is not None:
+        return size
+    return None
+
   def visualize(self, model_root, log_dir):
     for inputter in self.inputters:
       inputter.visualize(model_root, log_dir)
@@ -296,15 +350,18 @@ class ParallelInputter(MultiInputter):
     else:
       return tuple(inputter.input_signature() for inputter in self.inputters)
 
+  def _index_features(self, features, index):
+    if self.combine_features:
+      return misc.extract_prefixed_keys(features, "inputter_{}_".format(index))
+    else:
+      return features[index]
+
   def get_length(self, features, ignore_special_tokens=False):
-    lengths = []
-    for i, inputter in enumerate(self.inputters):
-      if self.combine_features:
-        sub_features = misc.extract_prefixed_keys(features, "inputter_{}_".format(i))
-      else:
-        sub_features = features[i]
-      lengths.append(inputter.get_length(
-          sub_features, ignore_special_tokens=ignore_special_tokens))
+    lengths = [
+        inputter.get_length(
+            self._index_features(features, i),
+            ignore_special_tokens=ignore_special_tokens)
+        for i, inputter in enumerate(self.inputters)]
     if self.reducer is None:
       return lengths
     else:
@@ -339,6 +396,24 @@ class ParallelInputter(MultiInputter):
             training=training)
       return tuple(features)
 
+  def keep_for_training(self, features, maximum_length=None):
+    if not isinstance(maximum_length, list):
+      maximum_length = [maximum_length]
+    # Unset maximum lengths are set to None (i.e. no constraint).
+    maximum_length += [None] * (len(self.inputters) - len(maximum_length))
+    constraints = []
+    for i, inputter in enumerate(self.inputters):
+      keep = inputter.keep_for_training(
+          self._index_features(features, i), maximum_length=maximum_length[i])
+      if isinstance(keep, bool):
+        if not keep:
+          return False
+        continue
+      constraints.append(keep)
+    if not constraints:
+      return True
+    return tf.reduce_all(constraints)
+
   def build(self, input_shape):
     if self.share_parameters:
       # When sharing parameters, build the first leaf inputter and then set
@@ -346,7 +421,7 @@ class ParallelInputter(MultiInputter):
       leaves = self.get_leaf_inputters()
       first, others = leaves[0], leaves[1:]
       first.build(input_shape)
-      for name, attr in first.__dict__.items():
+      for name, attr in first.__dict__.copy().items():
         if isinstance(attr, tf.Variable) or (isinstance(attr, tf.Module) and attr.variables):
           for inputter in others:
             setattr(inputter, name, attr)
@@ -357,13 +432,9 @@ class ParallelInputter(MultiInputter):
     super(ParallelInputter, self).build(input_shape)
 
   def call(self, features, training=None):
-    transformed = []
-    for i, inputter in enumerate(self.inputters):
-      if self.combine_features:
-        sub_features = misc.extract_prefixed_keys(features, "inputter_{}_".format(i))
-      else:
-        sub_features = features[i]
-      transformed.append(inputter(sub_features, training=training))
+    transformed = [
+        inputter(self._index_features(features, i), training=training)
+        for i, inputter in enumerate(self.inputters)]
     if self.reducer is not None:
       transformed = self.reducer(transformed)
     return transformed
@@ -428,38 +499,8 @@ class MixedInputter(MultiInputter):
     return outputs
 
 
-class ExampleInputter(ParallelInputter):
-  """An inputter that returns training examples (parallel features and labels)."""
-
-  def __init__(self, features_inputter, labels_inputter, share_parameters=False):
-    """Initializes this inputter.
-
-    Args:
-      features_inputter: An inputter producing the features (source).
-      labels_inputter: An inputter producing the labels (target).
-      share_parameters: Share the inputters parameters.
-    """
-    self.features_inputter = features_inputter
-    self.features_inputter.asset_prefix = "source"
-    self.labels_inputter = labels_inputter
-    self.labels_inputter.asset_prefix = "target"
-    super(ExampleInputter, self).__init__(
-        [self.features_inputter, self.labels_inputter],
-        share_parameters=share_parameters,
-        combine_features=False)
-
-  def make_inference_dataset(self,
-                             features_file,
-                             batch_size,
-                             length_bucket_width=None,
-                             num_threads=1,
-                             prefetch_buffer_size=None):
-    return self.features_inputter.make_inference_dataset(
-        features_file,
-        batch_size,
-        length_bucket_width=length_bucket_width,
-        num_threads=num_threads,
-        prefetch_buffer_size=prefetch_buffer_size)
+class ExampleInputterAdapter:
+  """Extends an inputter with methods to build evaluation and training datasets."""
 
   def make_evaluation_dataset(self,
                               features_file,
@@ -483,11 +524,18 @@ class ExampleInputter(ParallelInputter):
     See Also:
       :func:`opennmt.data.inference_pipeline`
     """
-    map_func = lambda *arg: self.make_features(element=arg, training=False)
-    dataset = self.make_dataset([features_file, labels_file], training=False)
+    if labels_file is not None:
+      data_files = [features_file, labels_file]
+    else:
+      data_files = features_file
+
+    map_fn = lambda *arg: self.make_features(element=misc.item_or_tuple(arg), training=False)
+    transform_fns = [lambda dataset: dataset.map(map_fn, num_parallel_calls=num_threads or 1)]
+
+    dataset = self.make_dataset(data_files, training=False)
     dataset = dataset.apply(dataset_util.inference_pipeline(
         batch_size,
-        process_fn=map_func,
+        transform_fns=transform_fns,
         num_threads=num_threads,
         prefetch_buffer_size=prefetch_buffer_size))
     return dataset
@@ -555,8 +603,25 @@ class ExampleInputter(ParallelInputter):
     See Also:
       :func:`opennmt.data.training_pipeline`
     """
-    map_func = lambda *arg: self.make_features(element=arg, training=True)
-    dataset = self.make_dataset([features_file, labels_file], training=True)
+    if labels_file is not None:
+      data_files = [features_file, labels_file]
+      maximum_length = [maximum_features_length, maximum_labels_length]
+      features_length_fn = self.features_inputter.get_length
+      labels_length_fn = self.labels_inputter.get_length
+    else:
+      data_files = features_file
+      maximum_length = maximum_features_length
+      features_length_fn = self.get_length
+      labels_length_fn = None
+
+    map_fn = lambda *arg: self.make_features(element=misc.item_or_tuple(arg), training=True)
+    filter_fn = lambda *arg: (
+        self.keep_for_training(misc.item_or_tuple(arg), maximum_length=maximum_length))
+    transform_fns = [
+        lambda dataset: dataset.map(map_fn, num_parallel_calls=num_threads or 4),
+        lambda dataset: dataset.filter(filter_fn)]
+
+    dataset = self.make_dataset(data_files, training=True)
     if weights is not None:
       dataset = (dataset, weights)
     dataset = dataset_util.training_pipeline(
@@ -564,17 +629,50 @@ class ExampleInputter(ParallelInputter):
         batch_type=batch_type,
         batch_multiplier=batch_multiplier,
         batch_size_multiple=batch_size_multiple,
-        process_fn=map_func,
+        transform_fns=transform_fns,
         length_bucket_width=length_bucket_width,
-        features_length_fn=self.features_inputter.get_length,
-        labels_length_fn=self.labels_inputter.get_length,
-        maximum_features_length=maximum_features_length,
-        maximum_labels_length=maximum_labels_length,
+        features_length_fn=features_length_fn,
+        labels_length_fn=labels_length_fn,
         single_pass=single_pass,
         num_shards=num_shards,
         shard_index=shard_index,
         num_threads=num_threads,
+        dataset_size=self.get_dataset_size(data_files),
         shuffle_buffer_size=shuffle_buffer_size,
         prefetch_buffer_size=prefetch_buffer_size,
         cardinality_multiple=cardinality_multiple)(dataset)
     return dataset
+
+
+class ExampleInputter(ParallelInputter, ExampleInputterAdapter):
+  """An inputter that returns training examples (parallel features and labels)."""
+
+  def __init__(self, features_inputter, labels_inputter, share_parameters=False):
+    """Initializes this inputter.
+
+    Args:
+      features_inputter: An inputter producing the features (source).
+      labels_inputter: An inputter producing the labels (target).
+      share_parameters: Share the inputters parameters.
+    """
+    self.features_inputter = features_inputter
+    self.features_inputter.asset_prefix = "source"
+    self.labels_inputter = labels_inputter
+    self.labels_inputter.asset_prefix = "target"
+    super(ExampleInputter, self).__init__(
+        [self.features_inputter, self.labels_inputter],
+        share_parameters=share_parameters,
+        combine_features=False)
+
+  def make_inference_dataset(self,
+                             features_file,
+                             batch_size,
+                             length_bucket_width=None,
+                             num_threads=1,
+                             prefetch_buffer_size=None):
+    return self.features_inputter.make_inference_dataset(
+        features_file,
+        batch_size,
+        length_bucket_width=length_bucket_width,
+        num_threads=num_threads,
+        prefetch_buffer_size=prefetch_buffer_size)
