@@ -18,7 +18,6 @@ class Model(tf.keras.layers.Layer):
     super(Model, self).__init__()
     self.examples_inputter = examples_inputter
     self.params = {}
-    self._frozen_layers = False
 
   @property
   def unsupervised(self):
@@ -35,21 +34,6 @@ class Model(tf.keras.layers.Layer):
     """The inputter producing labels."""
     return getattr(self.examples_inputter, "labels_inputter", None)
 
-  @property
-  def trainable_weights(self):
-    if not self._frozen_layers:
-      self._frozen_layers = True
-      freeze_layers = self.params.get("freeze_layers")
-      if freeze_layers:
-        if not isinstance(freeze_layers, list):
-          freeze_layers = [freeze_layers]
-        for layer_path in freeze_layers:
-          layer = misc.index_structure(self, layer_path)
-          layer.trainable = False
-        tf.get_logger().info("%d weights are frozen by the freeze_layers parameter" % (
-            len(self.non_trainable_weights)))
-    return super(Model, self).trainable_weights
-
   @property
   def ctranslate2_spec(self):
     """The equivalent CTranslate2 model specification."""
@@ -86,6 +70,14 @@ class Model(tf.keras.layers.Layer):
     self.examples_inputter.initialize(data_config)
 
   def build(self, input_shape):
+    freeze_layers = self.params.get("freeze_layers")
+    if freeze_layers:
+      if not isinstance(freeze_layers, list):
+        freeze_layers = [freeze_layers]
+      for layer_path in freeze_layers:
+        layer = misc.index_structure(self, layer_path)
+        layer.trainable = False
+        misc.set_dropout(layer, 0)  # Disable dropout in frozen layers.
     self.examples_inputter.build(input_shape)
     self.built = True
 
@@ -206,9 +198,13 @@ class Model(tf.keras.layers.Layer):
     """Returns the optimizer for this model.
 
     Returns:
-      A ``tf.keras.optimizers.Optimizer`` instance.
+      A ``tf.keras.optimizers.Optimizer`` instance or ``None`` if no optimizer
+      is configured.
     """
     params = self.params
+    optimizer_name = params.get("optimizer")
+    if optimizer_name is None:
+      return None
     learning_rate = tf.constant(params["learning_rate"], dtype=tf.float32)
     if params.get("decay_type") is not None:
       schedule_params = params.get("decay_params", {})
@@ -222,7 +218,7 @@ class Model(tf.keras.layers.Layer):
     if optimizer_params is None:
       optimizer_params = {}
     optimizer = optimizers.make_optimizer(
-        params["optimizer"], learning_rate, **optimizer_params)
+        optimizer_name, learning_rate, **optimizer_params)
     return optimizer
 
   def serve_function(self):
@@ -239,6 +235,9 @@ class Model(tf.keras.layers.Layer):
     @tf.function(input_signature=(input_signature,))
     def _run(features):
       features = self.features_inputter.make_features(features=features.copy())
+      if isinstance(features, (list, tuple)):
+        # Special case for unsupervised inputters that always return a tuple (features, labels).
+        features = features[0]
       _, predictions = self(features)
       return predictions
 
@@ -303,11 +302,11 @@ class Model(tf.keras.layers.Layer):
       raise ValueError("Transferring weights to another model type is not supported")
     if ignore_weights is None:
       ignore_weights = set()
-    ignore_weights_ref = set(weight.experimental_ref() for weight in ignore_weights)
+    ignore_weights_ref = set(weight.ref() for weight in ignore_weights)
     weights = self.weights
     new_weights = new_model.weights
     for weight, new_weight in zip(weights, new_weights):
-      if new_weight.experimental_ref() not in ignore_weights_ref:
+      if new_weight.ref() not in ignore_weights_ref:
         new_weight.assign(weight)
         if new_optimizer is not None and optimizer is not None:
           for slot_name in new_optimizer.get_slot_names():
