@@ -143,7 +143,6 @@ class Tokenizer(abc.ABC):
       tokens = [tf.compat.as_text(token) for token in tokens]
       return self._detokenize_string(tokens)
 
-  @tf.autograph.experimental.do_not_convert
   def _tokenize_tensor(self, text):
     """Tokenizes a tensor.
 
@@ -159,7 +158,7 @@ class Tokenizer(abc.ABC):
     def _python_wrapper(string_t):
       string = tf.compat.as_text(string_t.numpy())
       tokens = self._tokenize_string(string)
-      return tf.constant(tokens)
+      return tf.constant(tokens, dtype=tf.string)
     tokens = tf.py_function(_python_wrapper, [text], tf.string)
     tokens.set_shape([None])
     return tokens
@@ -182,7 +181,6 @@ class Tokenizer(abc.ABC):
         lambda x: tf.strings.reduce_join(self._tokenize_tensor(x), axis=0, separator=" "), text)
     return tf.strings.split(tokens, sep=" ")
 
-  @tf.autograph.experimental.do_not_convert
   def _detokenize_tensor(self, tokens):
     """Detokenizes tokens.
 
@@ -218,8 +216,7 @@ class Tokenizer(abc.ABC):
     return tf.map_fn(
         lambda x: self._detokenize_tensor(x[0][:x[1]]),
         (tokens, sequence_length),
-        dtype=tf.string,
-        back_prop=False)
+        dtype=tf.string)
 
   def _detokenize_ragged_tensor(self, tokens):
     """Detokenizes a batch of tokens as a ``tf.RaggedTensor``
