@@ -16,7 +16,7 @@ class Checkpoint(object):
     """Initializes the wrapper.
 
     Args:
-      model: A :class:`opennmt.models.model.Model` to save.
+      model: A :class:`opennmt.models.Model` to save.
       optimizer: The optimizer instance.
       model_dir: The directory where checkpoints will be saved. If not set, a
         temporary directory will be used.
@@ -35,6 +35,27 @@ class Checkpoint(object):
     self._checkpoint_manager = tf.train.CheckpointManager(
         self._checkpoint, model_dir, keep_checkpoint_max)
 
+  @classmethod
+  def from_config(cls, config, model, optimizer=None):
+    """Creates a checkpoint wrapper from the configuration.
+
+    Args:
+      config: The user configuration.
+      model: A :class:`opennmt.models.Model` to save.
+      optimizer: The optimizer instance.
+
+    Returns:
+      A :class:`openmt.utils.Checkpoint` instance.
+    """
+    train_config = config.get("train")
+    if train_config is None:
+      train_config = {}
+    return cls(
+        model,
+        optimizer=optimizer,
+        model_dir=config.get("model_dir"),
+        keep_checkpoint_max=train_config.get("keep_checkpoint_max", 8))
+
   @property
   def model(self):
     """The managed model."""
@@ -229,8 +250,8 @@ def _restore_v1_checkpoint(checkpoint_path, model, optimizer=None):
       del v1_structure[_V1_OPTIM_SCOPE]
       v1_structure = _merge_optimizer_slots(v1_structure, slots)
   mapping = model.map_v1_weights(v1_structure)
-  existing_variables = set(variable.experimental_ref() for variable in model.variables)
-  mapped_variables = set(variable.experimental_ref() for variable, _ in mapping)
+  existing_variables = set(variable.ref() for variable in model.variables)
+  mapped_variables = set(variable.ref() for variable, _ in mapping)
   missing_mapping = existing_variables.difference(mapped_variables)
   if missing_mapping:
     raise ValueError("The following variables were not mapped: %s" % (
