@@ -30,7 +30,8 @@ def get_devices(count=1, fallback_to_cpu=True):
   if not devices and fallback_to_cpu:
     devices = tf.config.list_logical_devices(device_type="CPU")
   if len(devices) < count:
-    raise ValueError("Requested %d devices but only %d are visible" % (count, len(devices)))
+    raise ValueError("Requested %d devices but only %d %s visible" % (
+        count, len(devices), "is" if len(devices) == 1 else "are"))
   return devices[0:count]
 
 def get_variables_name_mapping(root, root_key=None):
@@ -51,12 +52,12 @@ def get_variables_name_mapping(root, root_key=None):
   names_to_variables = {}
   for saveable_object in named_variables:
     variable = saveable_object.op
-    if not hasattr(variable, "experimental_ref"):  # Ignore non Tensor-like objects.
+    if not hasattr(variable, "ref"):  # Ignore non Tensor-like objects.
       continue
     name = saveable_object.name
     if root_key is not None:
       name = "%s/%s" % (root_key, name)
-    variables_to_names[variable.experimental_ref()] = name
+    variables_to_names[variable.ref()] = name
     names_to_variables[name] = variable
   return variables_to_names, names_to_variables
 
@@ -65,7 +66,7 @@ def get_variable_name(variable, root, model_key="model"):
   variables_to_names, _ = get_variables_name_mapping(root, root_key=model_key)
   # In case of a MirroredVariable, look up the primary variable
   variable = get_primary_variable(variable)
-  return variables_to_names.get(variable.experimental_ref())
+  return variables_to_names.get(variable.ref())
 
 def get_primary_variable(variable):
   """If :obj:`variable` is distributed, returns the primary component."""
@@ -134,13 +135,15 @@ def item_or_tuple(x):
   else:
     return x
 
-def count_lines(filename):
+def count_lines(filename, buffer_size=65536):
   """Returns the number of lines of the file :obj:`filename`."""
   with tf.io.gfile.GFile(filename, mode="rb") as f:
-    i = 0
-    for i, _ in enumerate(f):
-      pass
-    return i + 1
+    num_lines = 0
+    while True:
+      data = f.read(buffer_size)
+      if not data:
+        return num_lines
+      num_lines += data.count(b"\n")
 
 def is_gzip_file(filename):
   """Returns ``True`` if :obj:`filename` is a GZIP file."""
@@ -208,7 +211,7 @@ def set_dropout(root_layer, dropout):
   if not isinstance(root_layer, tf.Module):
     raise ValueError("Layer should be a tf.Module")
   for layer in (root_layer,) + root_layer.submodules:
-    for attr, value in layer.__dict__.items():
+    for attr, value in layer.__dict__.copy().items():
       if isinstance(value, tf.keras.layers.Dropout):
         value.rate = dropout
       elif "dropout" in attr:
@@ -308,24 +311,36 @@ class OrderRestorer(object):
     self._elements = {}
     self._heap = []
 
+  @property
+  def buffer_size(self):
+    """Number of elements waiting to be notified."""
+    return len(self._heap)
+
+  @property
+  def next_index(self):
+    """The next index to be notified."""
+    return self._next_index
+
   def _try_notify(self):
+    old_index = self._next_index
     while self._heap and self._heap[0] == self._next_index:
       index = heapq.heappop(self._heap)
       value = self._elements.pop(index)
       self._callback_fn(value)
       self._next_index += 1
+    return self._next_index != old_index
 
   def push(self, x):
     """Push event :obj:`x`."""
     index = self._index_fn(x)
     if index is None:
       self._callback_fn(x)
-      return
+      return True
     if index < self._next_index:
       raise ValueError("Event index %d was already notified" % index)
     self._elements[index] = x
     heapq.heappush(self._heap, index)
-    self._try_notify()
+    return self._try_notify()
 
 class ClassRegistry(object):
   """Helper class to create a registry of classes."""
